from pathlib import Path
import shutil
import subprocess
from multiprocessing import Pool
from . import File


class Transcoder:
    def __init__(self, encoder: Path, extension: str, input_root: Path, output_root: Path):
        self.encoder = encoder
        self.extension = extension
        self.input_root = input_root
        self.output_root = output_root

    def transcode(self):
        transcode_list = []
        for artist in self.input_root.iterdir():
            if artist.is_dir():
                for album in artist.iterdir():
                    if album.is_dir():
                        for file in album.iterdir():
                            if file.is_file():
                                transcode_list.append(File(file))
                    else:
                        print(f"Warning, skipping non-dir '{album}' found in artist '{artist.parts[-1]}'")
                        continue
            else:
                print(f"Warning, skipping non-dir '{artist}' found in root")
                continue
        self._transcode(transcode_list, self.encoder)

    def _transcode(self, transcode_list: list, encoder: Path, workers=16):
        worker_args = [(track, encoder) for track in transcode_list]
        with Pool(workers) as p:
            results = p.starmap_async(self.transcode_worker, worker_args)
            p.close()
            p.join()
            for result in results.get():
                print(result)

    def _transcode_single_thread(self, transcode_list: list, encoder: Path):
        worker_args = [(track, encoder) for track in transcode_list]
        for args in worker_args:
            print(self.transcode_worker(*args))


    def transcode_worker(self, track, encoder):
        if track.is_art:
            return self.copy_album_art(track)
        elif track.is_audio:
            return self.transcode_audio(track, encoder)
        else:
            return f"File {track.path} ignored"

    def copy_album_art(self, file: File):
        self.create_directories(file)
        output_path = file.output_file(self.output_root, file.extension[1:])
        if output_path.exists():
            return f"Skipped {output_path}"
        try:
            shutil.copy(file.path, output_path)
            return f"Successfully copied {output_path}"
        except FileExistsError:
            pass
        except Exception:
            return f"{file.path} failed to copy"

    def transcode_audio(self, track, encoder):
        enc_filename = encoder.parts[-1]
        track_output = track.output_file(self.output_root, self.extension)
        if track_output.exists():
            return f"Skipped {track_output}"
        self.create_directories(track)
        if enc_filename == "opusenc.exe":
            additional_args = ('--bitrate', '128', '--music')
            subprocess_args = (str(encoder),) + additional_args + (str(track.path), str(track_output))
        elif enc_filename == "qaac64.exe":
            additional_args = (str(track), '-o', str(track_output), '--threading')
            subprocess_args = (str(encoder),) + additional_args
        try:
            subprocess.run(subprocess_args, capture_output=True, text=True, check=True)
            return f"Transcoded '{track}' successfully."
        except Exception:
            return f"ERROR: Transcoding of '{track}' failed."

    def create_directories(self, track: File):
        if not track.artist_out(self.output_root).exists():
            try:
                track.artist_out(self.output_root).mkdir()
            except FileExistsError:
                pass
        if not track.album_out(self.output_root).exists():
            try:
                track.album_out(self.output_root).mkdir()
            except FileExistsError:
                pass
